from typing import Any, Dict
import json
from virtuals_acp.contract_clients.base_contract_client import BaseAcpContractClient
from virtuals_acp.models import OperationPayload


class ACPAccount:
    def __init__(
        self,
        contract_client: BaseAcpContractClient,
        id: int,
        client_address: str,
        provider_address: str,
        metadata: Dict[str, Any],
    ):
        self.contract_client = contract_client
        self.id = id
        self.client_address = client_address
        self.provider_address = provider_address
        self.metadata = metadata

    def update_metadata(self, metadata: Dict[str, Any]) -> OperationPayload:
        result = self.contract_client.update_account_metadata(
            self.id,
            json.dumps(metadata),
        )
        return result



# virtuals_acp/configs.py
from web3 import Web3

from typing import Literal, Optional, List, Dict, Any
from virtuals_acp.fare import Fare
from virtuals_acp.abis.abi import ACP_ABI
from virtuals_acp.abis.abi_v2 import ACP_V2_ABI
from virtuals_acp.models import X402Config

ChainEnv = Literal["base-sepolia", "base"]


class ACPContractConfig:
    def __init__(
        self,
        chain: ChainEnv,
        rpc_url: str,
        chain_id: int,
        contract_address: str,
        base_fare: Fare,
        alchemy_base_url: str,
        acp_api_url: str,
        alchemy_policy_id: str,
        abi: List[Dict[str, Any]],
        rpc_endpoint: Optional[str] = None,
        x402_config: Optional[X402Config] = None,
    ):
        self.chain = chain
        self.rpc_url = rpc_url
        self.chain_id = chain_id
        self.contract_address = contract_address
        self.base_fare = base_fare
        self.alchemy_base_url = alchemy_base_url
        self.acp_api_url = acp_api_url
        self.alchemy_policy_id = alchemy_policy_id
        self.abi = abi
        self.rpc_endpoint = rpc_endpoint
        self.x402_config = x402_config


BASE_SEPOLIA_CONFIG = ACPContractConfig(
    chain="base-sepolia",
    rpc_url="https://alchemy-proxy.virtuals.io/api/proxy/rpc",
    chain_id=84532,
    contract_address="0x8Db6B1c839Fc8f6bd35777E194677B67b4D51928",
    base_fare=Fare("0x036CbD53842c5426634e7929541eC2318f3dCF7e", 6),
    alchemy_base_url="https://alchemy-proxy.virtuals.io/api/proxy/wallet",
    alchemy_policy_id="186aaa4a-5f57-4156-83fb-e456365a8820",
    acp_api_url="https://acpx.virtuals.gg/api",
    abi=ACP_ABI,
)


BASE_SEPOLIA_ACP_X402_CONFIG = ACPContractConfig(
    chain="base-sepolia",
    rpc_url="https://alchemy-proxy.virtuals.io/api/proxy/rpc",
    chain_id=84532,
    contract_address="0x8Db6B1c839Fc8f6bd35777E194677B67b4D51928",
    base_fare=Fare("0x036CbD53842c5426634e7929541eC2318f3dCF7e", 6),
    alchemy_base_url="https://alchemy-proxy.virtuals.io/api/proxy/wallet",
    alchemy_policy_id="186aaa4a-5f57-4156-83fb-e456365a8820",
    acp_api_url="https://acpx.virtuals.gg/api",
    abi=ACP_ABI,
    x402_config=X402Config(
        url="https://dev-acp-x402.virtuals.io",
    ),
)


BASE_SEPOLIA_CONFIG_V2 = ACPContractConfig(
    chain="base-sepolia",
    rpc_url="https://alchemy-proxy.virtuals.io/api/proxy/rpc",
    chain_id=84532,
    contract_address="0xdf54E6Ed6cD1d0632d973ADECf96597b7e87893c",
    base_fare=Fare("0x036CbD53842c5426634e7929541eC2318f3dCF7e", 6),
    alchemy_base_url="https://alchemy-proxy.virtuals.io/api/proxy/wallet",
    alchemy_policy_id="186aaa4a-5f57-4156-83fb-e456365a8820",
    acp_api_url="https://acpx.virtuals.gg/api",
    abi=ACP_V2_ABI,
)


BASE_SEPOLIA_ACP_X402_CONFIG_V2 = ACPContractConfig(
    chain="base-sepolia",
    rpc_url="https://alchemy-proxy.virtuals.io/api/proxy/rpc",
    chain_id=84532,
    contract_address="0xdf54E6Ed6cD1d0632d973ADECf96597b7e87893c",
    base_fare=Fare("0x036CbD53842c5426634e7929541eC2318f3dCF7e", 6),
    alchemy_base_url="https://alchemy-proxy.virtuals.io/api/proxy/wallet",
    alchemy_policy_id="186aaa4a-5f57-4156-83fb-e456365a8820",
    acp_api_url="https://acpx.virtuals.gg/api",
    abi=ACP_V2_ABI,
    x402_config=X402Config(
        url="https://dev-acp-x402.virtuals.io",
    )
)


BASE_MAINNET_CONFIG = ACPContractConfig(
    chain="base",
    rpc_url="https://alchemy-proxy-prod.virtuals.io/api/proxy/rpc",
    chain_id=8453,
    contract_address="0x6a1FE26D54ab0d3E1e3168f2e0c0cDa5cC0A0A4A",
    base_fare=Fare("0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913", 6),
    alchemy_base_url="https://alchemy-proxy.virtuals.io/api/proxy/wallet",
    alchemy_policy_id="186aaa4a-5f57-4156-83fb-e456365a8820",
    acp_api_url="https://acpx.virtuals.io/api",
    abi=ACP_ABI,
)


BASE_MAINNET_ACP_X402_CONFIG = ACPContractConfig(
    chain="base",
    rpc_url="https://alchemy-proxy-prod.virtuals.io/api/proxy/rpc",
    chain_id=8453,
    contract_address="0x6a1FE26D54ab0d3E1e3168f2e0c0cDa5cC0A0A4A",
    base_fare=Fare("0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913", 6),
    alchemy_base_url="https://alchemy-proxy.virtuals.io/api/proxy/wallet",
    alchemy_policy_id="186aaa4a-5f57-4156-83fb-e456365a8820",
    acp_api_url="https://acpx.virtuals.io/api",
    abi=ACP_ABI,
    x402_config=X402Config(
        url="https://acp-x402.virtuals.io",
    ),
)



BASE_MAINNET_CONFIG_V2 = ACPContractConfig(
    chain="base",
    rpc_url="https://alchemy-proxy-prod.virtuals.io/api/proxy/rpc",
    chain_id=8453,
    contract_address="0xa6C9BA866992cfD7fd6460ba912bfa405adA9df0",
    base_fare=Fare("0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913", 6),
    alchemy_base_url="https://alchemy-proxy.virtuals.io/api/proxy/wallet",
    alchemy_policy_id="186aaa4a-5f57-4156-83fb-e456365a8820",
    acp_api_url="https://acpx.virtuals.io/api",
    abi=ACP_V2_ABI,
)


BASE_MAINNET_ACP_X402_CONFIG_V2 = ACPContractConfig(
    chain="base",
    rpc_url="https://alchemy-proxy-prod.virtuals.io/api/proxy/rpc",
    chain_id=8453,
    contract_address="0xa6C9BA866992cfD7fd6460ba912bfa405adA9df0",
    base_fare=Fare("0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913", 6),
    alchemy_base_url="https://alchemy-proxy.virtuals.io/api/proxy/wallet",
    alchemy_policy_id="186aaa4a-5f57-4156-83fb-e456365a8820",
    acp_api_url="https://acpx.virtuals.io/api",
    abi=ACP_V2_ABI,
    x402_config=X402Config(
        url="https://acp-x402.virtuals.io",
    ),
)


DEFAULT_CONFIG = BASE_MAINNET_CONFIG_V2
# Or: DEFAULT_CONFIG = BASE_SEPOLIA_CONFIG_V2



import functools
import json
import warnings
from typing import Optional, Type, Union, Dict, Any
import base64

from pydantic import ValidationError

from virtuals_acp.models import T


def get_txn_hash_from_response(response: Dict[str, Any]) -> Optional[str]:
    try:
        return response.get("receipts", [])[0].get("transactionHash")
    except (IndexError, AttributeError):
        print(f"Error getting transaction hash from response: {response}")
        return None


def try_parse_json_model(content: str, model: Type[T]) -> Optional[T]:
    try:
        return model.model_validate_json(content)
    except (json.JSONDecodeError, ValidationError):
        return None


def try_validate_model(data: dict, model: Type[T]) -> Optional[T]:
    try:
        return model.model_validate(data)
    except ValidationError:
        return None


def prepare_payload(payload: Union[str, Dict[str, Any]]) -> str:
    return payload if isinstance(payload, str) else json.dumps(payload)


def deprecated(reason: str = "This function is deprecated and should not be used."):
    """Decorator to mark functions or methods as deprecated."""

    def decorator(func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            warnings.warn(
                f"Call to deprecated function {func.__name__}: {reason}",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return wrapped

    return decorator


def safe_base64_encode(data: Union[str, bytes]) -> str:
    if isinstance(data, str):
        data_bytes = data.encode("utf-8")
    else:
        data_bytes = data
    return base64.b64encode(data_bytes).decode("utf-8")



from datetime import datetime
from typing import TYPE_CHECKING, Optional, Type, Dict, List, Any

from pydantic import BaseModel, ConfigDict

from virtuals_acp.models import (
    ACPJobPhase,
    MemoType,
    PayloadType,
    GenericPayload,
    T,
    ACPMemoStatus,
)
from virtuals_acp.utils import (
    try_parse_json_model,
    try_validate_model,
    get_txn_hash_from_response,
)

if TYPE_CHECKING:
    from virtuals_acp.contract_clients.base_contract_client import BaseAcpContractClient


class ACPMemo(BaseModel):
    contract_client: "BaseAcpContractClient"
    id: int
    type: MemoType
    content: str
    next_phase: ACPJobPhase
    status: ACPMemoStatus
    signed_reason: Optional[str] = None
    expiry: Optional[datetime] = None
    payable_details: Optional[Dict[str, Any]] = None
    txn_hash: Optional[str] = None
    signed_txn_hash: Optional[str] = None

    structured_content: Optional[GenericPayload] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def model_post_init(self, _):
        self.structured_content = try_parse_json_model(self.content, GenericPayload)

        if self.payable_details:
            self.payable_details["amount"] = int(self.payable_details["amount"])
            self.payable_details["feeAmount"] = int(self.payable_details["feeAmount"])

    def __str__(self):
        return f"AcpMemo({self.model_dump(exclude={'payable_details'})})"

    @property
    def payload_type(self) -> Optional[PayloadType]:
        if self.structured_content is not None:
            return self.structured_content.type

    def create(self, job_id: int, is_secured: bool = True):
        return self.contract_client.create_memo(
            job_id, self.content, self.type, is_secured, self.next_phase
        )

    def sign(self, approved: bool, reason: str | None = None) -> str | None:
        operation = self.contract_client.sign_memo(self.id, approved, reason)
        response = self.contract_client.handle_operation([operation])
        return get_txn_hash_from_response(response)



SINGLE_SIGNER_VALIDATION_MODULE_ABI = [
    { "inputs": [], "name": "InvalidSignatureType", "type": "error" },
    {
        "inputs": [],
        "name": "NotAuthorized",
        "type": "error",
    },
    { "inputs": [], "name": "NotImplemented", "type": "error" },
    {
        "inputs": [],
        "name": "UnexpectedDataPassed",
        "type": "error",
    },
    {
        "anonymous": True,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "account",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "uint32",
                "name": "entityId",
                "type": "uint32",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "newSigner",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "address",
                "name": "previousSigner",
                "type": "address",
            },
        ],
        "name": "SignerTransferred",
        "type": "event",
    },
    {
        "inputs": [],
        "name": "moduleId",
        "outputs": [{ "internalType": "string", "name": "", "type": "string" }],
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "inputs": [{ "internalType": "bytes", "name": "data", "type": "bytes" }],
        "name": "onInstall",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{ "internalType": "bytes", "name": "data", "type": "bytes" }],
        "name": "onUninstall",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            { "internalType": "address", "name": "account", "type": "address" },
            {
                "internalType": "bytes32",
                "name": "hash",
                "type": "bytes32",
            },
        ],
        "name": "replaySafeHash",
        "outputs": [{ "internalType": "bytes32", "name": "", "type": "bytes32" }],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            { "internalType": "uint32", "name": "entityId", "type": "uint32" },
            {
                "internalType": "address",
                "name": "account",
                "type": "address",
            },
        ],
        "name": "signers",
        "outputs": [{ "internalType": "address", "name": "", "type": "address" }],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{ "internalType": "bytes4", "name": "interfaceId", "type": "bytes4" }],
        "name": "supportsInterface",
        "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            { "internalType": "uint32", "name": "entityId", "type": "uint32" },
            {
                "internalType": "address",
                "name": "newSigner",
                "type": "address",
            },
        ],
        "name": "transferSigner",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            { "internalType": "address", "name": "account", "type": "address" },
            {
                "internalType": "uint32",
                "name": "entityId",
                "type": "uint32",
            },
            { "internalType": "address", "name": "sender", "type": "address" },
            {
                "internalType": "uint256",
                "name": "",
                "type": "uint256",
            },
            { "internalType": "bytes", "name": "", "type": "bytes" },
            {
                "internalType": "bytes",
                "name": "",
                "type": "bytes",
            },
        ],
        "name": "validateRuntime",
        "outputs": [],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            { "internalType": "address", "name": "account", "type": "address" },
            {
                "internalType": "uint32",
                "name": "entityId",
                "type": "uint32",
            },
            { "internalType": "address", "name": "", "type": "address" },
            {
                "internalType": "bytes32",
                "name": "digest",
                "type": "bytes32",
            },
            { "internalType": "bytes", "name": "signature", "type": "bytes" },
        ],
        "name": "validateSignature",
        "outputs": [{ "internalType": "bytes4", "name": "", "type": "bytes4" }],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {
                "internalType": "uint32",
                "name": "entityId",
                "type": "uint32",
            },
            {
                "components": [
                    { "internalType": "address", "name": "sender", "type": "address" },
                    {
                        "internalType": "uint256",
                        "name": "nonce",
                        "type": "uint256",
                    },
                    { "internalType": "bytes", "name": "initCode", "type": "bytes" },
                    {
                        "internalType": "bytes",
                        "name": "callData",
                        "type": "bytes",
                    },
                    {
                        "internalType": "bytes32",
                        "name": "accountGasLimits",
                        "type": "bytes32",
                    },
                    {
                        "internalType": "uint256",
                        "name": "preVerificationGas",
                        "type": "uint256",
                    },
                    { "internalType": "bytes32", "name": "gasFees", "type": "bytes32" },
                    {
                        "internalType": "bytes",
                        "name": "paymasterAndData",
                        "type": "bytes",
                    },
                    { "internalType": "bytes", "name": "signature", "type": "bytes" },
                ],
                "internalType": "struct PackedUserOperation",
                "name": "userOp",
                "type": "tuple",
            },
            { "internalType": "bytes32", "name": "userOpHash", "type": "bytes32" },
        ],
        "name": "validateUserOp",
        "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
        "stateMutability": "view",
        "type": "function",
    },
]



from dataclasses import dataclass, field
from typing import (
    Any,
    List,
    Optional,
    TYPE_CHECKING,
    Dict,
    Union,
    TypeVar,
    Generic,
    Literal,
)
from enum import Enum

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel

if TYPE_CHECKING:
    from virtuals_acp.job_offering import ACPJobOffering, ACPResourceOffering


class ACPMemoStatus(str, Enum):
    PENDING = "PENDING"
    APPROVED = "APPROVED"
    REJECTED = "REJECTED"
    EXPIRED = "EXPIRED"


class MemoType(int, Enum):
    MESSAGE = 0  # Text message
    CONTEXT_URL = 1  # URL for context
    IMAGE_URL = 2  # Image URL
    VOICE_URL = 3  # Voice/audio URL
    OBJECT_URL = 4  # Object/file URL
    TXHASH = 5  # Transaction hash reference
    PAYABLE_REQUEST = 6  # Payment request
    PAYABLE_TRANSFER = 7  # Direct payment transfer
    PAYABLE_TRANSFER_ESCROW = 8  # Escrowed payment transfer
    NOTIFICATION = 9  # Notification
    PAYABLE_NOTIFICATION = 10  # Payable notification


class ACPJobPhase(int, Enum):
    REQUEST = 0
    NEGOTIATION = 1
    TRANSACTION = 2
    EVALUATION = 3
    COMPLETED = 4
    REJECTED = 5
    EXPIRED = 6
    UNDEFINED = 999

    @classmethod
    def from_value(cls, value: str):
        try:
            return cls(value)
        except ValueError:
            return cls.UNDEFINED


class FeeType(int, Enum):
    NO_FEE = 0
    IMMEDIATE_FEE = 1
    DEFERRED_FEE = 2
    PERCENTAGE_FEE = 3


class PriceType(str, Enum):
    FIXED = "fixed"
    PERCENTAGE = "percentage"


class ACPAgentSort(str, Enum):
    SUCCESSFUL_JOB_COUNT = "successfulJobCount"
    SUCCESS_RATE = "successRate"
    UNIQUE_BUYER_COUNT = "uniqueBuyerCount"
    MINS_FROM_LAST_ONLINE = "minsFromLastOnlineTime"


class ACPGraduationStatus(str, Enum):
    GRADUATED = "graduated"
    NOT_GRADUATED = "not_graduated"
    ALL = "all"


class ACPOnlineStatus(str, Enum):
    ONLINE = "online"
    OFFLINE = "offline"
    ALL = "all"


DeliverablePayload = Union[str, Dict[str, Any]]
IDeliverable = DeliverablePayload  # Deprecated: use DeliverablePayload instead


@dataclass
class IACPAgent:
    id: int
    name: str
    description: str
    wallet_address: str  # Checksummed address
    job_offerings: List["ACPJobOffering"] = field(default_factory=list)
    resources: List["ACPResourceOffering"] = field(default_factory=list)
    twitter_handle: Optional[str] = None
    # Full fields from TS for completeness, though browse_agent returns a subset
    document_id: Optional[str] = None
    is_virtual_agent: Optional[bool] = None
    profile_pic: Optional[str] = None
    category: Optional[str] = None
    token_address: Optional[str] = None
    owner_address: Optional[str] = None
    cluster: Optional[str] = None
    symbol: Optional[str] = None
    virtual_agent_id: Optional[str] = None
    metrics: Optional[Dict[str, Any]] = None
    processing_time: Optional[str] = None
    contract_address: Optional[str] = None


class PayloadType(str, Enum):
    FUND_RESPONSE = "fund_response"
    OPEN_POSITION = "open_position"
    CLOSE_POSITION = "close_position"
    CLOSE_PARTIAL_POSITION = "close_partial_position"
    POSITION_FULFILLED = "position_fulfilled"
    CLOSE_JOB_AND_WITHDRAW = "close_job_and_withdraw"
    UNFULFILLED_POSITION = "unfulfilled_position"


T = TypeVar("T", bound=BaseModel)


class PayloadModel(BaseModel):
    model_config = ConfigDict(alias_generator=to_camel, validate_by_name=True)

    # JSON-friendly payload fields when using model_dump and model_dump_json
    def model_dump(self, *args, **kwargs):
        kwargs.setdefault("by_alias", True)
        return super().model_dump(*args, **kwargs)

    def model_dump_json(self, *args, **kwargs):
        kwargs.setdefault("by_alias", True)
        return super().model_dump_json(*args, **kwargs)

    def __str__(self):
        return f"{self.__class__.__name__}({self.model_dump(by_alias=False)})"


class GenericPayload(PayloadModel, Generic[T]):
    type: PayloadType
    data: T | List[T]


class RequestPayload(PayloadModel):
    name: Optional[str] = None
    requirement: Optional[Union[str, Dict[str, Any]]] = None
    service_name: Optional[str] = None
    service_requirement: Optional[Dict[str, Any]] = None
    price_type: PriceType = PriceType.FIXED
    price_value: Optional[float] = None
    model_config = ConfigDict(extra="allow")


class FundResponsePayload(PayloadModel):
    reporting_api_endpoint: str
    wallet_address: Optional[str] = None


class TPSLConfig(PayloadModel):
    price: Optional[float] = None
    percentage: Optional[float] = None


class PositionDirection(str, Enum):
    LONG = "long"
    SHORT = "short"


class OpenPositionPayload(PayloadModel):
    symbol: str
    amount: float
    chain: Optional[str] = None
    contract_address: Optional[str] = None
    direction: Optional[PositionDirection] = None
    tp: TPSLConfig
    sl: TPSLConfig


class UpdateTPSLConfig(PayloadModel):
    amount_percentage: Optional[float] = None


class UpdatePositionPayload(PayloadModel):
    symbol: str
    contract_address: Optional[str] = None
    tp: Optional[UpdateTPSLConfig] = None
    sl: Optional[UpdateTPSLConfig] = None


class ClosePositionPayload(PayloadModel):
    position_id: int
    amount: float


class PositionFulfilledPayload(PayloadModel):
    symbol: str
    amount: float
    contract_address: str
    type: Literal["TP", "SL", "CLOSE"]
    pnl: float
    entry_price: float
    exit_price: float


class UnfulfilledPositionPayload(PayloadModel):
    symbol: str
    amount: float
    contract_address: str
    type: Literal["ERROR", "PARTIAL"]
    reason: Optional[str] = None


class CloseJobAndWithdrawPayload(PayloadModel):
    message: str


class RequestClosePositionPayload(PayloadModel):
    position_id: int


class AcpJobX402PaymentDetails(PayloadModel):
    is_x402: bool
    is_budget_received: bool


class X402Config(PayloadModel):
    url: str


class X402RequirementExtra(PayloadModel):
    name: str
    version: str


class X402Requirement(PayloadModel):
    scheme: str
    network: str
    maxAmountRequired: str
    resource: str
    description: str
    mimeType: str
    payTo: str  # Address as str
    maxTimeoutSeconds: int
    asset: str  # Address as str
    extra: X402RequirementExtra
    outputSchema: Any


class X402PayableRequirements(PayloadModel):
    x402Version: int
    error: str
    accepts: List[X402Requirement]


class X402PayableRequest(PayloadModel):
    to: str  # Address as str
    value: int
    maxTimeoutSeconds: int
    asset: str  # Address as str


class X402Payment(PayloadModel):
    encodedPayment: str
    message: Dict[str, Any]
    signature: str


class X402PaymentPayload(PayloadModel):
    x402_version: int
    scheme: str
    network: str
    payload: Dict[str, Any]


class OperationPayload(PayloadModel):
    data: str  # Should start with '0x'
    to: str  # Address as str
    value: Optional[int] = None


class OffChainJob(PayloadModel):
    id: int
    documentId: str
    txHash: str  # Address as str
    clientId: int
    providerId: int
    budget: float
    createdAt: str
    updatedAt: str
    publishedAt: str
    locale: Optional[str] = None
    clientAddress: str
    providerAddress: str
    evaluators: List[str]
    budgetTxHash: Optional[str] = None
    phase: str  # AcpJobPhases as str (define separately if available)
    agentIdPair: str
    onChainJobId: str
    summary: str
    userOpHash: Optional[str] = None
    amountClaimed: float
    context: Optional[Dict[str, Any]] = None
    expiry: str
    refundRetryTimes: int
    additionalFees: float
    budgetTokenAddress: str
    budgetUSD: float
    amountClaimedUSD: Optional[float] = None
    additionalFeesUSD: Optional[float] = None
    contractAddress: str
    accountId: Optional[int] = None
    x402Nonce: str


class X402PaymentResponse(PayloadModel):
    isPaymentRequired: bool
    data: X402PayableRequirements



class ACPError(Exception):
    """Base exception for ACP client errors."""

    pass


class ACPApiError(ACPError):
    """Raised for errors from the ACP API."""

    pass


class ACPContractError(ACPError):
    """Raised for errors interacting with the ACP smart contract."""

    pass


class TransactionFailedError(ACPContractError):
    """Raised when a blockchain transaction fails."""

    pass



FIAT_TOKEN_V2_ABI = [
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "owner",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "spender",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256",
            },
        ],
        "name": "Approval",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "authorizer",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "bytes32",
                "name": "nonce",
                "type": "bytes32",
            },
        ],
        "name": "AuthorizationCanceled",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "authorizer",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "bytes32",
                "name": "nonce",
                "type": "bytes32",
            },
        ],
        "name": "AuthorizationUsed",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "_account",
                "type": "address",
            },
        ],
        "name": "Blacklisted",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "newBlacklister",
                "type": "address",
            },
        ],
        "name": "BlacklisterChanged",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "burner",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256",
            },
        ],
        "name": "Burn",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "newMasterMinter",
                "type": "address",
            },
        ],
        "name": "MasterMinterChanged",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "minter",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount",
                "type": "uint256",
            },
        ],
        "name": "Mint",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "minter",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "minterAllowedAmount",
                "type": "uint256",
            },
        ],
        "name": "MinterConfigured",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "oldMinter",
                "type": "address",
            },
        ],
        "name": "MinterRemoved",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "previousOwner",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "address",
                "name": "newOwner",
                "type": "address",
            },
        ],
        "name": "OwnershipTransferred",
        "type": "event",
    },
    {"anonymous": False, "inputs": [], "name": "Pause", "type": "event"},
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "newAddress",
                "type": "address",
            },
        ],
        "name": "PauserChanged",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "newRescuer",
                "type": "address",
            },
        ],
        "name": "RescuerChanged",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "from",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "to",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "value",
                "type": "uint256",
            },
        ],
        "name": "Transfer",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "_account",
                "type": "address",
            },
        ],
        "name": "UnBlacklisted",
        "type": "event",
    },
    {"anonymous": False, "inputs": [], "name": "Unpause", "type": "event"},
    {
        "inputs": [],
        "name": "CANCEL_AUTHORIZATION_TYPEHASH",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "DOMAIN_SEPARATOR",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "PERMIT_TYPEHASH",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "RECEIVE_WITH_AUTHORIZATION_TYPEHASH",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "TRANSFER_WITH_AUTHORIZATION_TYPEHASH",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "owner", "type": "address"},
            {"internalType": "address", "name": "spender", "type": "address"},
        ],
        "name": "allowance",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
        ],
        "name": "approve",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "authorizer", "type": "address"},
            {"internalType": "bytes32", "name": "nonce", "type": "bytes32"},
        ],
        "name": "authorizationState",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "account", "type": "address"}],
        "name": "balanceOf",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "_account", "type": "address"}],
        "name": "blacklist",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "blacklister",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "uint256", "name": "_amount", "type": "uint256"}],
        "name": "burn",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "authorizer", "type": "address"},
            {"internalType": "bytes32", "name": "nonce", "type": "bytes32"},
            {"internalType": "uint8", "name": "v", "type": "uint8"},
            {"internalType": "bytes32", "name": "r", "type": "bytes32"},
            {"internalType": "bytes32", "name": "s", "type": "bytes32"},
        ],
        "name": "cancelAuthorization",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "authorizer", "type": "address"},
            {"internalType": "bytes32", "name": "nonce", "type": "bytes32"},
            {"internalType": "bytes", "name": "signature", "type": "bytes"},
        ],
        "name": "cancelAuthorization",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "minter", "type": "address"},
            {
                "internalType": "uint256",
                "name": "minterAllowedAmount",
                "type": "uint256",
            },
        ],
        "name": "configureMinter",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "currency",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "decimals",
        "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "decrement", "type": "uint256"},
        ],
        "name": "decreaseAllowance",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "increment", "type": "uint256"},
        ],
        "name": "increaseAllowance",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "string", "name": "tokenName", "type": "string"},
            {"internalType": "string", "name": "tokenSymbol", "type": "string"},
            {"internalType": "string", "name": "tokenCurrency", "type": "string"},
            {"internalType": "uint8", "name": "tokenDecimals", "type": "uint8"},
            {"internalType": "address", "name": "newMasterMinter", "type": "address"},
            {"internalType": "address", "name": "newPauser", "type": "address"},
            {"internalType": "address", "name": "newBlacklister", "type": "address"},
            {"internalType": "address", "name": "newOwner", "type": "address"},
        ],
        "name": "initialize",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "string", "name": "newName", "type": "string"}],
        "name": "initializeV2",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "lostAndFound", "type": "address"},
        ],
        "name": "initializeV2_1",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "internalType": "address[]",
                "name": "accountsToBlacklist",
                "type": "address[]",
            },
            {"internalType": "string", "name": "newSymbol", "type": "string"},
        ],
        "name": "initializeV2_2",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "_account", "type": "address"}],
        "name": "isBlacklisted",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "account", "type": "address"}],
        "name": "isMinter",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "masterMinter",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_to", "type": "address"},
            {"internalType": "uint256", "name": "_amount", "type": "uint256"},
        ],
        "name": "mint",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "minter", "type": "address"}],
        "name": "minterAllowance",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "name",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "owner", "type": "address"}],
        "name": "nonces",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "owner",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "pause",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "paused",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "pauser",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "owner", "type": "address"},
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
            {"internalType": "uint256", "name": "deadline", "type": "uint256"},
            {"internalType": "bytes", "name": "signature", "type": "bytes"},
        ],
        "name": "permit",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "owner", "type": "address"},
            {"internalType": "address", "name": "spender", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
            {"internalType": "uint256", "name": "deadline", "type": "uint256"},
            {"internalType": "uint8", "name": "v", "type": "uint8"},
            {"internalType": "bytes32", "name": "r", "type": "bytes32"},
            {"internalType": "bytes32", "name": "s", "type": "bytes32"},
        ],
        "name": "permit",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "from", "type": "address"},
            {"internalType": "address", "name": "to", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
            {"internalType": "uint256", "name": "validAfter", "type": "uint256"},
            {"internalType": "uint256", "name": "validBefore", "type": "uint256"},
            {"internalType": "bytes32", "name": "nonce", "type": "bytes32"},
            {"internalType": "bytes", "name": "signature", "type": "bytes"},
        ],
        "name": "receiveWithAuthorization",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "from", "type": "address"},
            {"internalType": "address", "name": "to", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
            {"internalType": "uint256", "name": "validAfter", "type": "uint256"},
            {"internalType": "uint256", "name": "validBefore", "type": "uint256"},
            {"internalType": "bytes32", "name": "nonce", "type": "bytes32"},
            {"internalType": "uint8", "name": "v", "type": "uint8"},
            {"internalType": "bytes32", "name": "r", "type": "bytes32"},
            {"internalType": "bytes32", "name": "s", "type": "bytes32"},
        ],
        "name": "receiveWithAuthorization",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "minter", "type": "address"}],
        "name": "removeMinter",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "internalType": "contract IERC20",
                "name": "tokenContract",
                "type": "address",
            },
            {"internalType": "address", "name": "to", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "rescueERC20",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "rescuer",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "symbol",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "totalSupply",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "to", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
        ],
        "name": "transfer",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "from", "type": "address"},
            {"internalType": "address", "name": "to", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
        ],
        "name": "transferFrom",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "newOwner", "type": "address"}],
        "name": "transferOwnership",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "from", "type": "address"},
            {"internalType": "address", "name": "to", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
            {"internalType": "uint256", "name": "validAfter", "type": "uint256"},
            {"internalType": "uint256", "name": "validBefore", "type": "uint256"},
            {"internalType": "bytes32", "name": "nonce", "type": "bytes32"},
            {"internalType": "bytes", "name": "signature", "type": "bytes"},
        ],
        "name": "transferWithAuthorization",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "from", "type": "address"},
            {"internalType": "address", "name": "to", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
            {"internalType": "uint256", "name": "validAfter", "type": "uint256"},
            {"internalType": "uint256", "name": "validBefore", "type": "uint256"},
            {"internalType": "bytes32", "name": "nonce", "type": "bytes32"},
            {"internalType": "uint8", "name": "v", "type": "uint8"},
            {"internalType": "bytes32", "name": "r", "type": "bytes32"},
            {"internalType": "bytes32", "name": "s", "type": "bytes32"},
        ],
        "name": "transferWithAuthorization",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "_account", "type": "address"}],
        "name": "unBlacklist",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "unpause",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_newBlacklister", "type": "address"},
        ],
        "name": "updateBlacklister",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_newMasterMinter", "type": "address"},
        ],
        "name": "updateMasterMinter",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_newPauser", "type": "address"}
        ],
        "name": "updatePauser",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "newRescuer", "type": "address"}
        ],
        "name": "updateRescuer",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "version",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "pure",
        "type": "function",
    },
]
